/*
 * App.java
 *
 * Created: April 7, 2009, 12:55 PM
 * Author: Stephen Whitmore
 *
 */
package irrtum;

import irrtum.map.*;
import irrtum.game.*;
import irrtum.engine.*;
import java.awt.Color;
import org.lwjgl.opengl.*;
import org.lwjgl.util.glu.*;
import org.lwjgl.input.*;
import javax.swing.*;

/**
 * Responsible for creating the game instance itself and maintaining it.
 */
public class Client
{
    static Model mdl;
    public static final String VERSION = "0.1";
    
    static DisplayMode dMode;
    public static final int TARGET_UPS = 60;
    public static final int WIDTH = 800, HEIGHT = 600, DEPTH = 24;
    static int frameCount, updateCount, ups, fps;
    static long ticks=0, lastFrame;
    static double updateSum;
    static boolean isRunning;
    static Camera gameCamera;
    static Map map;

    public static void crash(String msg)
    {
        JOptionPane.showMessageDialog(null, msg, "irrtum "+VERSION+" has crashed!", JOptionPane.OK_OPTION);
        System.exit(1);
    }
    
    public static void run()
    {
        long lastSecond = lastFrame = getTime();
        
        // Create LWJGL window
        if(!initDisplay(WIDTH, HEIGHT, 60))
        {
            if(!initDisplay(WIDTH, HEIGHT, -1))
            {
                crash("No available display mode found. Do you have hardware-acceleration enabled?");
            }
        }

        // Load some testing resources
        ResourceManager.addImage("blocks.png", Color.PINK, "Blocks");
        ResourceManager.addImage("billboard.png", new Color(255,0,255), "RiotGuy");
        ResourceManager.addModel("wall.3ds", "Wall");
        ResourceManager.addModel("floor.3ds", "Floor");

        // Init map
        map = new Map();

        // Init fog
        GL11.glLoadIdentity();
        GL11.glEnable(GL11.GL_FOG);
        GL11.glFogi(GL11.GL_FOG_MODE, GL11.GL_LINEAR);
        GL11.glHint(GL11.GL_FOG_HINT, GL11.GL_NICEST);
        Weather.setFog(new Vector3(0.7f,0.7f,0.7f), -20.0f, 1500.0f);

        // Init sun
        //GL11.glEnable(GL11.GL_LIGHTING);
        GL11.glEnable(GL11.GL_LIGHT0);
        Sun.change(new Vector3(0.75f,0.75f,0.75f), new Vector3(1.75f,1.75f,1.75f), new Vector3(0.84f,0.24f,-0.48f));

        // Init camera
        gameCamera = new Camera(2000);
        gameCamera.setLook(new Vector3(0,0,1));
        gameCamera.setPos(new Vector3(0,10,0));

        BillboardEntity riotGuy = new BillboardEntity();
        riotGuy.setTexture(ResourceManager.getImage("RiotGuy").getTexture());
        riotGuy.setPos(new Vector3(16,16,16));
        EntityEngine.addEntity(riotGuy);
        
        // Main loop
        isRunning = true;
        lastFrame = getTime();
        lastSecond = getTime();
        final float interval = (1000.0f / (float)TARGET_UPS);
        while(isRunning)
        {
            // Render current session
            render();
            frameCount++;
	    
            // Update current session
            if(getTime()-lastFrame >= interval)
            {
                updateSum += getTime() - lastFrame;
                lastFrame = getTime();
                int numUpdates = (int)(updateSum / interval);
                for(int i=0; i < numUpdates; i++)
                {
                    update();
                    updateCount++;
                    ticks++;
                }
                updateSum -= interval*numUpdates;
            }
	    
            // Check if a second has passed to recalculate UPS and FPS
            if( getTime() - lastSecond >= 1000 )
            {
                lastSecond = getTime();
                ups = updateCount;
                fps = frameCount;
                updateCount = frameCount = 0;
                Display.setTitle("FPS: "+fps+"    UPS: "+ups);
            }
        }
        
        ResourceManager.destroyImages();
        Display.destroy();
        
        System.exit(0);
    }
    
    public static void close()
    {
        isRunning = false;
    }

    static float p=0;
    public static void render()
    {
        GL11.glClear( GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT );
		
        map.render(gameCamera);
        
        EntityEngine.render(gameCamera);
        //ParticleEngine.render();
        
        //HUD.render();
        //GuiManager.render();
        
        //drawMiniMap();

        Display.update();
    }
    
    public static void update()
    {
        EntityEngine.update();

        Input.update();
    	
    	if(Input.isKeyHit(Keyboard.KEY_ESCAPE) || Display.isCloseRequested())
    		close();
        if(Input.isKeyDown(Keyboard.KEY_W))
            gameCamera.pos.set( gameCamera.pos.add(gameCamera.look.scale(3)) );
        if(Input.isKeyDown(Keyboard.KEY_S))
            gameCamera.pos.set( gameCamera.pos.add(gameCamera.look.scale(-2)) );
        if(Input.isKeyDown(Keyboard.KEY_A))
        {
            Vector3 strafeDir = gameCamera.look.rotateY(-90.0f);
            strafeDir.setY(0);
            gameCamera.pos.set( gameCamera.pos.add(strafeDir.norm().scale(3.0f)));
        }
        if(Input.isKeyDown(Keyboard.KEY_D))
        {
            Vector3 strafeDir = gameCamera.look.rotateY(90.0f);
            strafeDir.setY(0);
            gameCamera.pos.set( gameCamera.pos.add(strafeDir.norm().scale(3.0f)));
        }

        Vector2 mouseMickey = Input.getMouseMickey();
        //System.out.println("Mickey: "+mouseMickey.toString());
        gameCamera.rotateY(mouseMickey.getX());
        gameCamera.rotateX(-mouseMickey.getY());

        //System.out.println("Look: "+gameCamera.look.toString());
    }
    
    private static boolean initDisplay(int w, int h, int freq)
    {
        try
        {
            DisplayMode[] modes = Display.getAvailableDisplayModes();

            for(int i=0; i < modes.length; i++)
            {
                int bpp = modes[i].getBitsPerPixel();
                if( modes[i].getWidth() == w && modes[i].getHeight() == h && (freq == -1 || modes[i].getFrequency() == freq) )
                {
                    if(dMode == null || bpp > dMode.getBitsPerPixel() )
                    {
                        if(!(modes[i].getFrequency() == 60 && freq == -1))
                            dMode = modes[i];
                    }
                    //System.out.println("Found mode: "+modes[i].toString());
                }
            }
        }
        catch(Exception e)
        {
            return false;
        }
                
        if(dMode == null)
        {
            return false;
        }
        
        try
        {
            //Display.setFullscreen( Config.isFullscreen() );
            Display.setFullscreen(false);
            Display.setDisplayMode(dMode);
            Display.setVSyncEnabled(false);
            Display.create();
        }
        catch(Exception e)
        {
            return false;
        }

        // Initialize OpenGL settings
        GL11.glMatrixMode(GL11.GL_PROJECTION);
        GL11.glLoadIdentity();
        GLU.gluPerspective(60.0f, w/(float)h, 1.0f, 2000.0f);
        GL11.glMatrixMode(GL11.GL_MODELVIEW);
        GL11.glLoadIdentity();
        GL11.glEnable(GL11.GL_TEXTURE_2D);
        GL11.glEnable(GL11.GL_ALPHA_TEST);
        GL11.glAlphaFunc(GL11.GL_GREATER, 0.02f);
        GL11.glEnable(GL11.GL_DEPTH_TEST);
        GL11.glClearColor(0.527f, 0.804f, 0.976f, 1.0f);
        return true;
    }
    
    public static long getTime()
    {
	//return (Sys.getTime() * 1000) / Sys.getTimerResolution();
        return System.currentTimeMillis();
    }
    
    public static long getTicks()
    {
        return ticks;
    }
    
    public static int getFPS()
    {
        return fps;
    }
    
    public static int getUPS()
    {
        return ups;
    }
    
    public static void main(String[] args)
    {
        run();
    }
}
