/*
 * BitmapFont.java
 *
 * Created: July 15, 2007, 10:34 AM
 * Author: Stephen Whitmore
 *
 */
package irrtum.engine;

import java.awt.*;
import java.io.*;
import java.util.*;
import org.lwjgl.opengl.*;

/**
 * Uses the AngelCode bitmap font file format to generate OpenGL textures in order to
 * render text to the screen.
 *
 */
public class BitmapFont
{
    Texture texture;
    BitmapGlyph[] glyphs;
    int lineHeight;
    
    public BitmapFont(String fontName)
    {
	glyphs = new BitmapGlyph[256];
	
	texture = Texture.loadFromFile("fonts/"+fontName+".png", Color.BLACK, true, false);
	
	try
        {
            InputStream instr = Thread.currentThread().getContextClassLoader().getResourceAsStream("fonts/"+fontName+".fnt");
	    BufferedReader br = new BufferedReader( new InputStreamReader( instr ) );
//	    BufferedReader br = new BufferedReader( new FileReader( "Fonts/"+fontName+".fnt" ) );
	    Scanner scan;
            
            scan = new Scanner(br.readLine());
	    lineHeight = scan.nextInt();	    
	    
	    String line;
	    while( (line=br.readLine()) != null )
	    {
		int id = getNumVal("id", line);
		int x = getNumVal("x", line);
		int y = getNumVal("y", line);
		int w = getNumVal("width", line);
		int h = getNumVal("height", line);
		int xo = getNumVal("xoffset", line);
		int yo = getNumVal("yoffset", line);
		int xa = getNumVal("xadvance", line);
		
		BitmapGlyph gl = new BitmapGlyph(this, x,y,w,h,xo,yo,xa);
		glyphs[id] = gl;
		//System.out.println("Loaded glyph: \""+(char)id+"\" ("+x+","+y+","+w+","+h+","+xo+","+yo+","+xa+")");
		scan.close();
	    }
	}
	catch(Exception e)
	{
	    System.out.println("BitmapFont Load Error ("+fontName+"): "+e.toString());
	}
    }
    
    private int getNumVal(String prop, String txt) throws IOException
    {
	int index = txt.indexOf(prop);
	if(index == -1)
	{
            throw new IOException("Could not properly parse Font info.");
	    //System.out.println("NumVal Prop not found, \""+prop+"\".");
	}
	index += prop.length() + 1;
	String res="";
	while( !Character.isWhitespace( txt.charAt(index) ) )
	{
	    res += txt.charAt(index);
	    index++;
	}
	return Integer.parseInt(res);
    }
    
    public void draw(float x, float y, String text)
    {
        Color col = Color.WHITE;
        draw(x,y,1,1,col,col,100,false,null,text);
    }
    public void draw(float x, float y, float z, boolean world, String text)
    {
        Color col = Color.WHITE;
        draw(x,y,1,1,col,col,z,world,null,text);
    }
    public void draw(float x, float y, float z, boolean world, Color col, String text)
    {
        draw(x,y,1,1,col,col,z,world,null,text);
    }
    public void draw(float x, float y, float scaleX, float scaleY, float z, boolean world, String text)
    {
        Color col = Color.WHITE;
        draw(x,y,scaleX,scaleY,col,col,z,world,null,text);
    }
    public void draw(float x, float y, float scaleX, float scaleY, Color col, float z, boolean world, String text)
    {
        draw(x,y,scaleX,scaleY,col,col,z,world,null,text);
    }
    // OPTIMIZATION: Convert into display lists (so that only one glCallLists() per string is needed)
    public void draw(float x, float y, float scaleX, float scaleY, Color topCol, Color botCol, float z, boolean world, Color outline, String text)
    {
	if(texture == null)
	    return;
        
        // Cap to integer coordinates to prevent subpixel ugliness
        x = (int)x;
        y = (int)y;

	if(outline != null)
	{
	    //draw(x+1,y+1,scaleX,scaleY,outline,outline,z-0.1f,world,null,text);
	    //draw(x-1,y-1,scaleX,scaleY,outline,outline,z-0.1f,world,null,text);
	    //draw(x+1,y-1,scaleX,scaleY,outline,outline,z-0.1f,world,null,text);
	    //draw(x-1,y+1,scaleX,scaleY,outline,outline,z-0.1f,world,null,text);
	}
        
	GL11.glPushMatrix();
	GL11.glLoadIdentity();

    // TODO: Transform based on a given camera?
    // ...
    
	GL11.glTranslatef(x, y, z);
	
	texture.bind();
	GL11.glScalef(scaleX, scaleY, 1);
        
        Image.setBlendMode(Image.BLEND_ALPHA);
	        
        GL11.glBegin(GL11.GL_QUADS);
	int xo = 0;
	for(int i=0; i < text.length(); i++)
	{
	    char ch = text.charAt(i);
            if(ch >= 0 && ch < glyphs.length)
            {
                BitmapGlyph gl = glyphs[ch];
                if(gl == null)
                    continue;
                if(outline != null)
                {
                    gl.draw(xo-1,0,outline,outline,-0.01f);
                    gl.draw(xo,-1,outline,outline,-0.01f);
                    gl.draw(xo+1,0,outline,outline,-0.01f);
                    gl.draw(xo,1,outline,outline,-0.01f);
                }
                gl.draw(xo, 0, topCol, botCol, 0);
                xo += gl.getXAdvance();
            }
	}
	GL11.glEnd();
	
	GL11.glPopMatrix();
	GL11.glColor4f(1,1,1,1);
    }
    
    public Texture getTexture()
    {
	return texture;
    }
    
    public int getTextWidth(String text)
    {
        if(texture == null)
            return 0;
        
        int w = 0;
	for(int i=0; i < text.length(); i++)
	{
	    char ch = text.charAt(i);
            if(ch < 0 || ch >= glyphs.length)
                continue;
	    BitmapGlyph gl = glyphs[ch];
	    if(gl == null)
		continue;
            w += gl.getXAdvance();
	}
        return w;
    }
    public int getHeight()
    {
        return lineHeight;
    }
}


class BitmapGlyph
{
    int x, y, width, height, xOffset, yOffset, xAdvance;
    BitmapFont parent;
    
    public BitmapGlyph(BitmapFont par, int bx, int by, int w, int h, int xo, int yo, int xa)
    {
	parent = par;
	x = bx;
	y = by;
	width = w;
	height = h;
	xOffset = xo;
	yOffset = yo;
	xAdvance = xa;
    }
    
    public int getXAdvance()
    {
	return xAdvance;
    }
    
    public void draw(int xo, int yo, Color top, Color bot, float zo)
    {
	float wdiv = parent.getTexture().getWidth();
	float hdiv = parent.getTexture().getHeight();
	float tx = (x / wdiv);
	float ty = (y / hdiv);
	float tw = (width / wdiv);
	float th = (height / hdiv);
	
        GL11.glColor4ub((byte)top.getRed(), (byte)top.getGreen(), (byte)top.getBlue(), (byte)top.getAlpha());

        GL11.glTexCoord2f(tx,ty);
        GL11.glVertex3f(xo+xOffset,yo+yOffset,zo);

        GL11.glTexCoord2f(tx+tw,ty);
        GL11.glVertex3f(xo+xOffset+width,yo+yOffset,zo);

        GL11.glColor4ub((byte)bot.getRed(), (byte)bot.getGreen(), (byte)bot.getBlue(), (byte)bot.getAlpha());

        GL11.glTexCoord2f(tx+tw,ty+th);
        GL11.glVertex3f(xo+xOffset+width,yo+yOffset+height,zo);

        GL11.glTexCoord2f(tx,ty+th);
        GL11.glVertex3f(xo+xOffset,yo+yOffset+height,zo);
    }
}