/*
 * Image.java
 *
 * Created on July 5, 2007, 12:46 PM
 *
 */
package irrtum.engine;

import java.awt.*;
import org.lwjgl.opengl.*;

/**
 * A more complex object that wraps around a Texture, containing various
 * attributes and methods for more sophisticated image rendering.
 *
 * @author Stephen Whitmore
 *
 */
public class Image
{
    String name;
    Texture texture;
    int frameWidth, frameHeight;
    
    // Image blending modes
    public static final int BLEND_NONE      = 0;
    public static final int BLEND_ALPHA     = 1;
    public static final int BLEND_ADD       = 2;
    
    
    public Image(String imgName)
    {
        name = imgName;
        texture = null;
        frameWidth = frameHeight = 0;
    }
    public Image(String fileName, Color colourKey, String imgName)
    {
        texture = Texture.loadFromFile(fileName, colourKey, false, false);
        name = imgName;
        frameWidth = texture.getRealWidth();
        frameHeight = texture.getRealHeight();
    }
    public Image(Texture tex, String name)
    {
        texture = tex;
        name = name;
        frameWidth = texture.getRealWidth();
        frameHeight = texture.getRealHeight();
    }

    public int getFrameHeight()
    {
        return frameHeight;
    }

    public int getFrameWidth()
    {
        return frameWidth;
    }

    public String getName()
    {
        return name;
    }

    public Texture getTexture()
    {
        return texture;
    }

    public void setFrameHeight(int frameHeight)
    {
        this.frameHeight = frameHeight;
    }

    public void setFrameWidth(int frameWidth)
    {
        this.frameWidth = frameWidth;
    }
    public void setFrameSize(int w, int h)
    {
	frameWidth = w;
	frameHeight = h;
    }
    
    public int getWidth()
    {
        return getTexture().getWidth();
    }
    public int getHeight()
    {
        return getTexture().getHeight();
    }

    public void setName(String name)
    {
        this.name = name;
    }

    public void setTexture(Texture texture)
    {
        this.texture = texture;
    }
    
    public int getNumHorzFrames()
    {
        return (texture.getRealWidth() / frameWidth);
    }
    public int getNumVertFrames()
    {
        return (texture.getRealHeight() / frameHeight);
    }
    public int getNumFrames()
    {
        return getNumHorzFrames() * getNumVertFrames();
    }
        
    public void draw(Vector2 loc)
    {
        Color col = Color.WHITE;
        Vector2 centre = Vector2.create(0.5f,0.5f);
        Vector2 scale = Vector2.create(1,1);
        draw(loc,centre,0,scale,col,col,col,col,BLEND_NONE,0,0,false);
    }
    public void draw(Vector2 loc, float z, boolean world)
    {
        Color col = Color.WHITE;
        Vector2 centre = Vector2.create(0.5f,0.5f);
        Vector2 scale = Vector2.create(1,1);
        draw(loc,centre,0,scale,col,col,col,col,BLEND_NONE,z,0,world);
    }
    public void draw(Vector2 loc, float z, int frame, boolean world)
    {
        Color col = Color.WHITE;
        Vector2 centre = Vector2.create(0.5f,0.5f);
        Vector2 scale = Vector2.create(1,1);
        draw(loc,centre,0,scale,col,col,col,col,BLEND_NONE,z,frame,world);
    }
    public void draw(Vector2 loc, Vector2 centre, float rot, Vector2 scale, float z, int frame, boolean world)
    {
        Color col = Color.WHITE;
        draw(loc,centre,rot,scale,col,col,col,col,BLEND_NONE,z,frame,world);        
    }
    public void draw(Vector2 loc, Vector2 centre, float rot, Vector2 scale, Color col, int blend, float z, int frame, boolean world)
    {
        draw(loc,centre,rot,scale,col,col,col,col,blend,z,frame,world);
    }
    public void draw(Vector2 loc, Vector2 centre, float rot, Vector2 scale, Color c1, Color c2, Color c3, Color c4, int blend, float z, int frame, boolean world)
    {
        boolean dep = GL11.glIsEnabled(GL11.GL_DEPTH_TEST);
        if(z == -1)
        {
            GL11.glDisable(GL11.GL_DEPTH_TEST);
            GL11.glDepthMask(false);
        }
        boolean deallocC=false, deallocS=false;
        if(centre == null)
        {
            centre = Vector2.create(0.5f,0.5f);
            deallocC = true;
        }
        if(scale == null)
        {
            scale = Vector2.create(1,1);
            deallocS = true;
        }
        if(c1 == null)
            c1 = c2 = c3 = c4 = Color.WHITE;
        
        // Frame-locating/drawing values
        float frw = getFrameWidth() / (float)texture.getWidth();
        float frh = getFrameHeight() / (float)texture.getHeight();
        float frw_steps = frw * (frame % getNumHorzFrames());
        float frh_steps = frh * (frame / getNumHorzFrames());
        
        // Offsets for centre/scaling/etc.
        float cox = getFrameWidth() * centre.getX();
        float coy = getFrameHeight() * centre.getY();
        float rs = getFrameWidth() - cox;
        float bs = getFrameHeight() - coy;
        float wr = texture.getRealWidth() / (float)texture.getWidth();
        float wh = texture.getRealHeight() / (float)texture.getHeight();
        
        // Preparing OpenGL states for image drawing
        GL11.glPushMatrix();

        // TODO: Orthographic/screen-space transform
        // ...

        GL11.glTranslatef(loc.getX()+cox, loc.getY()+coy, z);
        GL11.glScalef(scale.getX(), scale.getY(), 1);
        GL11.glRotatef(rot, 0,0,1);        
        texture.bind();
        
        // Setting the blend mode
        Image.setBlendMode(blend);
        
        // Draw the image
        GL11.glBegin(GL11.GL_QUADS);
            GL11.glColor4ub((byte)c1.getRed(), (byte)c1.getGreen(), (byte)c1.getBlue(), (byte)c1.getAlpha());
            GL11.glTexCoord2f(frw_steps,frh_steps);
            GL11.glVertex2f(-cox,-coy);
            
            GL11.glColor4ub((byte)c2.getRed(), (byte)c2.getGreen(), (byte)c2.getBlue(), (byte)c2.getAlpha());
            GL11.glTexCoord2f(frw_steps+frw,frh_steps);
            GL11.glVertex2f(rs,-coy);

            GL11.glColor4ub((byte)c3.getRed(), (byte)c3.getGreen(), (byte)c3.getBlue(), (byte)c3.getAlpha());
            GL11.glTexCoord2f(frw_steps+frw,frh_steps+frh);
            GL11.glVertex2f(rs,bs);

            GL11.glColor4ub((byte)c4.getRed(), (byte)c4.getGreen(), (byte)c4.getBlue(), (byte)c4.getAlpha());
            GL11.glTexCoord2f(frw_steps,frh_steps+frh);
            GL11.glVertex2f(-cox,bs);
        GL11.glEnd();
        
        // Clean up changed state information
        GL11.glPopMatrix();
        GL11.glColor4f(1,1,1,1);
        
        if(dep)
        {
            GL11.glEnable(GL11.GL_DEPTH_TEST);
            GL11.glDepthMask(true);
        }
    }
    
    public void useLinearFiltering(boolean lin)
    {
        if(texture == null)
            return;
        
        texture.bind();
        if(lin)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        }
        else
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
        }
    }
    
    public static void setBlendMode(int blend)
    {
        GL11.glDisable(GL11.GL_BLEND);
        
        if(blend != BLEND_NONE)
            GL11.glEnable(GL11.GL_BLEND);
        
        if(blend == BLEND_ADD)
            GL11.glBlendFunc(GL11.GL_ONE, GL11.GL_ONE);
        else
            GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
    }
}
