/*
 * Model.java
 *
 * Created: April 7, 2009, 6:51 PM
 * Author: Stephen Whitmore
 *
 */
package irrtum.engine;

import java.io.*;
import java.nio.*;
import java.net.URL;
import org.lwjgl.opengl.*;

/**
 *
 * @author stephen
 */
public class Model
{
    private final static int MAIN_CHUNK = 0x4D4D;
    private final static int EDITOR_CHUNK = 0x3D3D;
    private final static int OBJECT_CHUNK = 0x4000;
    private final static int TRIANGULAR_MESH = 0x4100;
    private final static int VERTICES_LIST = 0x4110;
    private final static int FACES_DESCRIPTION = 0x4120;
//    private final static int FACES_MATERIAL = 0x4130;
    private final static int MAPPING_COORDINATES_LIST = 0x4140;

    private int numVertices, numFaces, numCoords;
    private float[][] vertices, normals, texcoords;
    private short[][] faces;
    private FloatBuffer vArray, tArray;
    private boolean loaded;
    private float scale=1.0f;

    public Model()
    {
        loaded = false;
    }

    public boolean loadFromFile(String fname)
    {
        try
        {
            URL url = Thread.currentThread().getContextClassLoader().getResource(fname);
            if(url == null) throw new Exception();
            File file = new File(url.getFile());
            DataInputStream str = new DataInputStream( new FileInputStream( new File(fname) ) );
            long fileSize = file.length();

            byte[] raw = new byte[(int)fileSize];
            str.readFully(raw);
            ByteBuffer data = ByteBuffer.wrap(raw).order(ByteOrder.LITTLE_ENDIAN);
            str.close();

            int chunkType=0;
            int chunkLen=0;

            while(data.position() < data.limit())
            {
                //System.out.println("File Pos: "+data.position());
                chunkType = data.getShort();
                chunkLen = data.getInt();
                //System.out.println("Chunk Type: "+Integer.toHexString(chunkType));
                //System.out.println("Chunk Length: "+chunkLen+" ("+Integer.toHexString(chunkLen)+")");

                switch(chunkType)
                {
                    case MAIN_CHUNK:
                    {
                        //System.out.println("Main chunk");
                        break;
                    }

                    case EDITOR_CHUNK:
                    {
                        //System.out.println("Editor chunk");
                        break;
                    }

                    case OBJECT_CHUNK:
                    {
                        String name="";
                        byte ch=1;
                        while(ch != 0)
                        {
                            ch = data.get();
                            //System.out.println("Got char: "+(char)ch);
                            if(ch != 0)
                                name = name + ch;
                        }
                        //System.out.println("Object chunk ("+name+")");
                        break;
                    }

                    case TRIANGULAR_MESH:
                    {
                        //System.out.println("Triangular mesh");
                        break;
                    }

                    case VERTICES_LIST:
                    {
                        short numVerts = data.getShort();
                        //System.out.println("Vertices list ("+numVerts+")");

                        numVertices = numVerts;
                        vertices = new float[numVertices][3];

                        for(short i=0; i < numVerts; i++)
                        {
                            float x,y,z;
                            x = data.getFloat();
                            y = data.getFloat();
                            z = data.getFloat();
                            vertices[i][0] = x;
                            vertices[i][1] = y;
                            vertices[i][2] = z;
                            //System.out.println("Vertex ("+x+", "+y+", "+z+")");
                        }
                        break;
                    }

                    case FACES_DESCRIPTION:
                    {
                        short numPolys = data.getShort();
                        //System.out.println("Faces Description ("+numPolys+")");

                        numFaces = numPolys;
                        faces = new short[numFaces][3];
                        normals = new float[numFaces][3];

                        for(short i=0; i < numPolys; i++)
                        {
                            short xi, yi, zi, flags;
                            xi = data.getShort();
                            yi = data.getShort();
                            zi = data.getShort();
                            flags = data.getShort();
                            faces[i][0] = xi;
                            faces[i][1] = yi;
                            faces[i][2] = zi;

                            // Calculate normal for face
                            float x1 = vertices[yi][0] - vertices[xi][0];
                            float y1 = vertices[yi][1] - vertices[xi][1];
                            float z1 = vertices[yi][2] - vertices[xi][2];
                            float x2 = vertices[zi][0] - vertices[yi][0];
                            float y2 = vertices[zi][1] - vertices[yi][1];
                            float z2 = vertices[zi][2] - vertices[yi][2];
                            normals[i][0] = y1 * z2 - z1 * y2;
                            normals[i][1] = z1 * x2 - x1 * z2;
                            normals[i][2] = x1 * y2 - y1 * x2;

                            // HACK: Flip normal
                            float mag = - (float)Math.sqrt( normals[i][0]*normals[i][0] + normals[i][1]*normals[i][1] + normals[i][2]*normals[i][2] );
                            normals[i][0] /= mag;
                            normals[i][1] /= mag;
                            normals[i][2] /= mag;
                            //System.out.println("Created face ("+xi+", "+yi+", "+zi+")");
                            //std:cout << "Normal: (" << normals[i][0] << ", " << normals[i][1] << ", " << normals[i][2] << ")\n";
                            //std::cout << "Polygon Index (" << xi << ", " << yi << ", " << zi << ") (Flags=" << flags << ")\n";
                        }
                        break;
                    }

                    case MAPPING_COORDINATES_LIST:
                    {
                        //System.out.println("Mapping Coords List");
                        short coord_count = data.getShort();
                        numCoords = coord_count;

                        texcoords = new float[numVertices][2];

                        for(short i=0; i < numCoords; i++)
                        {
                            float u, v;
                            u = data.getFloat();
                            v = data.getFloat();
                            texcoords[i][0] = u;
                            texcoords[i][1] = v;
                            //System.out.println("Texture Coords ("+u+", "+v+")");
                        }
                        break;
                    }

                    default:
                    {
                        //System.out.println("Unknown chunk.");
                        data.position(data.position() + chunkLen - 6);
                        //fseek(file, chunkLen-6, SEEK_CUR);
                        break;
                    }
                }
            }

            if(null == vertices || null == faces || null == texcoords) { return false; }
            //System.out.println("Model data read successfully.");

            // Build vertex arrays
            
            vArray = ByteBuffer.allocateDirect( numFaces * 3 * 3 * Float.SIZE).asFloatBuffer();
            tArray = ByteBuffer.allocateDirect( numFaces * 3 * 2 * Float.SIZE).asFloatBuffer();
            vArray.rewind();
            tArray.rewind();
            for(int i=0; i < numFaces; i++)
            {
                float[] v1 = vertices[ faces[i][0] ];
                float[] v2 = vertices[ faces[i][1] ];
                float[] v3 = vertices[ faces[i][2] ];
                float[] t1 = texcoords[ faces[i][0] ];
                float[] t2 = texcoords[ faces[i][1] ];
                float[] t3 = texcoords[ faces[i][2] ];
                float[] n = normals[ i ];

                vArray.put(i*3*3+0, v1[0]);
                vArray.put(i*3*3+1, v1[2]);
                vArray.put(i*3*3+2, v1[1]);
                vArray.put(i*3*3+3, v2[0]);
                vArray.put(i*3*3+4, v2[2]);
                vArray.put(i*3*3+5, v2[1]);
                vArray.put(i*3*3+6, v3[0]);
                vArray.put(i*3*3+7, v3[2]);
                vArray.put(i*3*3+8, v3[1]);

                tArray.put(i*3*2+0, t1[0]);
                tArray.put(i*3*2+1, t1[1]);
                tArray.put(i*3*2+2, t2[0]);
                tArray.put(i*3*2+3, t2[1]);
                tArray.put(i*3*2+4, t3[0]);
                tArray.put(i*3*2+5, t3[1]);
            }
            //System.out.println("Vertex arrays created.");
        }
        catch(Exception e)
        {
            // TODO: This is bad.
            System.out.println("Model loading failed.");
            e.printStackTrace();
            return false;
        }
        
        loaded = true;

        return true;
    }
/*
    float* Model::calculateVertexLighting(Vec3 pos, float rot)
    {
        float * cArray = new float[numFaces*3*3];

        for(int i=0; i < numFaces; i++)
        {
            float *v1 = vertices[ faces[i][0] ];
            float *v2 = vertices[ faces[i][1] ];
            float *v3 = vertices[ faces[i][2] ];
            float *t1 = texcoords[ faces[i][0] ];
            float *t2 = texcoords[ faces[i][1] ];
            float *t3 = texcoords[ faces[i][2] ];
            float *n = normals[ i ];
            Vec3 p, col;

            Vec3 norm = Vec3( n[0], n[1], n[2] ).rotateZ(rot);

            p = Vec3(v1[0], v1[1], v1[2]).scale(scale).rotateZ(rot) + pos;
            col = gMap->getLightingAt( p, norm, true );
            cArray[i*3*3+0] = col.x;
            cArray[i*3*3+1] = col.y;
            cArray[i*3*3+2] = col.z;

            p = Vec3(v2[0], v2[1], v2[2]).scale(scale).rotateZ(rot) + pos;
            col = gMap->getLightingAt( p, norm, true );
            cArray[i*3*3+3] = col.x;
            cArray[i*3*3+4] = col.y;
            cArray[i*3*3+5] = col.z;

            p = Vec3(v3[0], v3[1], v3[2]).scale(scale).rotateZ(rot) + pos;
            col = gMap->getLightingAt( p, norm, true );
            cArray[i*3*3+6] = col.x;
            cArray[i*3*3+7] = col.y;
            cArray[i*3*3+8] = col.z;
        }

        return cArray;
    }
*/
    public void draw()
    {
        GL11.glPushMatrix();
//        GL11.glScalef(scale,scale,scale);
//        GL11.glColor3f(1,1,1);
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        /*
        GL11.glEnableClientState(GL11.GL_VERTEX_ARRAY);
        GL11.glEnableClientState(GL11.GL_TEXTURE_COORD_ARRAY);
        //if(cArray) glEnableClientState(GL_COLOR_ARRAY);

        GL11.glVertexPointer(3, 0, vArray);
        GL11.glTexCoordPointer(2, 0, tArray);
        //if(cArray) glColorPointer(3, GL_FLOAT, 0, cArray);

        GL11.glDrawArrays(GL11.GL_TRIANGLES, 0, numFaces*3);

        GL11.glDisableClientState(GL11.GL_VERTEX_ARRAY);
        GL11.glDisableClientState(GL11.GL_TEXTURE_COORD_ARRAY);
        //if(cArray) glDisableClientState(GL_COLOR_ARRAY);
         */
    //*
        GL11.glColor3f(1,1,1);
        GL11.glBegin(GL11.GL_TRIANGLES);
        for(int i=0; i < numFaces; i++)
        {
            float []v1 = vertices[ faces[i][0] ];
            float []v2 = vertices[ faces[i][1] ];
            float []v3 = vertices[ faces[i][2] ];
            float []t1 = texcoords[ faces[i][0] ];
            float []t2 = texcoords[ faces[i][1] ];
            float []t3 = texcoords[ faces[i][2] ];
            float []n = normals[ i ];
            Vector3 p, col;

            //Vector3 norm = Vec3( n[0], n[1], n[2] );
            Vector3 p1 = new Vector3(v1[0], v1[1], v1[2]);
            Vector3 p2 = new Vector3(v2[0], v2[1], v2[2]);
            Vector3 p3 = new Vector3(v3[0], v3[1], v3[2]);

            GL11.glNormal3f(n[0], n[1], n[2]);
            GL11.glTexCoord2f(t1[0], t1[1]);
            GL11.glVertex3f(p1.x, p1.y, p1.z);
            GL11.glTexCoord2f(t2[0], t2[1]);
            GL11.glVertex3f(p2.x, p2.y, p2.z);
            GL11.glTexCoord2f(t3[0], t3[1]);
            GL11.glVertex3f(p3.x, p3.y, p3.z);
        }
        GL11.glEnd();
    
        GL11.glPopMatrix();
    }

    public boolean isLoaded()
    {
        return loaded;
    }
}
