/*
 * ResourceManager.java
 *
 * Created: July 2, 2007, 1:06 PM
 * Author: Stephen Whitmore
 *
 */
package irrtum.engine;

import java.util.*;
import java.util.zip.*;
import java.io.*;
import java.awt.image.*;
import java.awt.*;
import javax.imageio.*;
import irrtum.*;

/**
 * Handle the creation, storage, retrieval, and destruction of game resources.
 * (eg. textures, sounds, music)
 */
public class ResourceManager
{
    static HashMap<String,Image> images;
    static HashMap<String,Model> models;
//    static HashMap<String,SoundEffect> sounds;
//    static HashMap<String,MusicTrack> music;
    
    static
    {
        images = new HashMap<String,Image>();
        models = new HashMap<String,Model>();
    }
    
    public static Model addModel(String filename, String name)
    {
        Model mdl = new Model();
        boolean res = mdl.loadFromFile(filename);
        if(res)
        {
            addModel(mdl, name);
            return mdl;
        }
        else
            return null;
    }
    public static void addModel(Model mdl, String name)
    {
        models.put(name, mdl);
        Log.write("Added model resource, \""+name+"\"", LogChannel.Info);
    }
    public static Model getModel(String name)
    {
        Model mdl = models.get(name);
        if(mdl == null)
        {
            Log.write("Could not locate Resource (Model) requested: "+name+".", LogChannel.Warning);
        }
        return mdl;
    }

    public static Image addImage(String filename, java.awt.Color ckey, String name)
    {
        Texture tex = Texture.loadFromFile(filename,ckey, false, false);
        if(tex == null)
            Client.crash("Failed to load image, \""+filename+"\".");
        Image img = new Image(tex,name);
        addImage(img, name);
        return img;
    }
    public static Image addImage(String filename, String name)
    {
        return addImage(filename, null, name);
    }
    public static void addImage(Image img, String name)
    {
        images.put(name, img);
        Log.write("Added image resource, \""+name+"\"", LogChannel.Info);
    }
    public static void addImage(Image img)
    {
        addImage(img, img.getName());
    }
    public static Image getImage(String name)
    {
        Image img = images.get(name);
        if(img == null)
        {
            Log.write("Could not locate Resource (Image) requested: "+name+".", LogChannel.Warning);
        }
        return img;
    }
    
    public static void destroyImages()
    {
        for(Image img : images.values())
        {
            if(img.getTexture() != null)
                img.getTexture().destroy();
        }
    }
    
    public static boolean loadResPak(String fileName)
    {
	try
	{
        InputStream instr = Thread.currentThread().getContextClassLoader().getResourceAsStream(fileName);
        if(instr == null)
        {
            Log.write("Unable to load ResPak, \""+fileName+"\". Fatal error.", LogChannel.Error);
            return false;
        }
        
        int cached=0;
	    ZipInputStream zin = new ZipInputStream(instr);
//	    ZipInputStream zin = new ZipInputStream(new FileInputStream(fileName));
	    HashMap<String,BufferedImage> imageMap = new HashMap<String,BufferedImage>();
	    ArrayList<String> imgList = new ArrayList<String>();
	    Scanner scanner=null;
	    
	    ZipEntry zip;
	    while( (zip=zin.getNextEntry()) != null )
	    {
            String name = zip.getName();
            //String imgName = name.substring(0, name.length()-4);
            String ext = name.substring( name.length()-3 );
            if(ext.toLowerCase().equals("png") || ext.toLowerCase().equals("bmp"))
            {
                DataInputStream str = new DataInputStream(zin);
                BufferedImage img = ImageIO.read(str);
                imageMap.put(name, img);
            }
            else if(ext.equals("lst"))
            {
                BufferedReader br = new BufferedReader( new InputStreamReader( zin ) );
                String str=null;
                while( (str=br.readLine()) != null )
                imgList.add(str);
                //System.out.println("Found ImgList entry.");
            }
            else
            {
                System.out.println("Skipped Entry: "+name);
                continue;
            }
            zin.closeEntry();
	    }
	    zin.close();
	    
	    
	    for(String str : imgList)
	    {
                if(str.startsWith("#"))
                    continue;
                if(str.trim().length() < 3)
                    continue;
		scanner = new Scanner(str);
		String fName = scanner.next();
		//String iName = scanner.next();
		String iName = fName.substring( 0, fName.length()-4 );
                //System.out.println(iName);
                boolean clamp = Integer.parseInt( scanner.next() ) == 1 ? true : false;
                boolean filter = Integer.parseInt( scanner.next() ) == 1 ? true : false;
                int fw = Integer.parseInt( scanner.next() );
		int fh = Integer.parseInt( scanner.next() );
		String colKey = scanner.next();
		int r=-1, g=-1, b=-1;
		if(colKey.equals("Y"))
		{
		    r = Integer.parseInt( scanner.next() );
		    g = Integer.parseInt( scanner.next() );
		    b = Integer.parseInt( scanner.next() );
		}

		BufferedImage img = imageMap.get(fName);
                if(img == null)
                {
                    Log.write("Failed to find requested image in ResPak: \""+fName+"\".", LogChannel.Error);
                    return false;
                }
		Texture tex;
		if(r != -1)
		    tex = Texture.loadFromData(img, new Color(r,g,b), filter, clamp);
		else
		    tex = Texture.loadFromData(img, null, filter, clamp);
		if(tex == null)
		{
		    System.out.println("Failed to load Entry: "+fName);
		    continue;
		}

		Image image = new Image(tex, iName);
		if(fw != 0 && fh != 0)
		    image.setFrameSize(fw,fh);
		addImage(image, iName);
                image.setName(iName);
		//System.out.println("Loaded Image: "+fName+" ("+iName+")");
                //Log.write("Loaded Image: "+fName+" ("+iName+")", LogChannel.Debug);
                cached++;
	    }
            Log.write("ResPak ("+fileName+") successfully loaded. "+cached+" images cached. ("+imageMap.values().size()+" loaded)", LogChannel.Info);
	    //System.out.println(imageMap.values().size()+" images loaded.");
            return true;
	}
	catch(IOException e)
	{
	    Log.write("IOException while loading ResPak: "+e.toString(), LogChannel.Error);
            return false;
	}
    }
}
