/*
 * Texture.java
 *
 * Created: July 2, 2007, 1:11 PM
 * Author: Stephen Whitmore
 *
 */
package irrtum.engine;

import java.nio.*;
import java.io.*;
import java.net.URL;
import java.awt.image.*;
import java.awt.*;
import javax.imageio.*;
import org.lwjgl.*;
import org.lwjgl.opengl.*;
import org.lwjgl.util.glu.*;

/**
 * A low-level object representing an OpenGL hardware texture.
 */
public class Texture
{
    static int currentID;    
    int textureID;
    int realWidth, realHeight;
    int texWidth, texHeight;
    int depth;
    Color[] texData;
    boolean hasAlpha;
    
    
    // The next power of two that is larger than the given number.
    // (For ensuring all textures are a power-of-two.
    static int nextPOT(int n)
    {
        int v = 1;
        while(v < n)
            v *= 2;
        return v;
    }
    

    public int getID()
    {
	return textureID;
    }
    
    public int getWidth()
    {
	return texWidth;	
    }
    
    public int getHeight()
    {
	return texHeight;
    }
    
    public int getRealWidth()
    {
        return realWidth;
    }
    
    public int getRealHeight()
    {
        return realHeight;
    }
    
    public void bind()
    {
	if(currentID != textureID)
	{
	    GL11.glBindTexture(GL11.GL_TEXTURE_2D, textureID);
	    currentID = textureID;
	}
    }
    
    // Have OpenGL unallocate the memory used for the texture
    public void destroy()
    {
	IntBuffer tmp = BufferUtils.createIntBuffer(1);
	tmp.put(0, textureID);
	GL11.glDeleteTextures(tmp);
    }
    
    // Load a texture from a file -- applying colour-keying if needed
    public static Texture loadFromFile(String filename)
    {
	return loadFromFile(filename, null, false, false);
    }
    public static Texture loadFromFile(String filename, Color colourKey, boolean linearFilter, boolean clampEdges)
    {
	BufferedImage img;
	
	try
	{
            URL url = Thread.currentThread().getContextClassLoader().getResource(filename);
            img = ImageIO.read(url);
            if(img == null)
            {
                throw new IOException();
            }
            else
                return loadFromData(img, colourKey, linearFilter, clampEdges);
	}
	catch(Exception e)
	{
	    System.out.println("ERROR: Could not load image, \""+filename+"\".");
	    return null;
	}	
    }
    
    public static Texture loadFromData(BufferedImage img, Color colourKey, boolean linearFilter, boolean clampEdges)
    {
        Texture tex=null;
        
        tex = new Texture();
        tex.hasAlpha = (colourKey != null);
        tex.realWidth = img.getWidth();
        tex.realHeight = img.getHeight();
        tex.texWidth = nextPOT(tex.realWidth);
        tex.texHeight = nextPOT(tex.realHeight);
        tex.depth = img.getType() == BufferedImage.TYPE_INT_RGB ? 3 : 4;
        //System.out.println("rW="+tex.realWidth+"  rH="+tex.realHeight+"  tW="+tex.texWidth+"  tH="+tex.texHeight);
        //System.out.println(tex.realWidth+","+tex.realHeight+","+tex.depth);
        
        int[] data = new int[img.getWidth()*img.getHeight()*tex.depth];
        data = img.getRGB(0,0, img.getWidth(),img.getHeight(), data, 0, img.getWidth());
        IntBuffer tmp = BufferUtils.createIntBuffer(1);
        int size = tex.texWidth * tex.texHeight * 4;
        tex.texData = new Color[tex.realWidth * tex.realHeight];
        ByteBuffer imgData = BufferUtils.createByteBuffer(size);
        
        for(int y=0; y < tex.realHeight; y++)
        {
            for(int x=0; x < tex.realWidth; x++)
            {
                int i = (y*tex.texWidth + x)*4;
                int j = (y*tex.realWidth + x);
                Color col = new Color(data[j]);
                int alpha = col.getAlpha();
                if(colourKey != null && col.equals(colourKey))
                {
                    alpha = 0;
                    col = Color.BLACK;
                }
                imgData.put(i+3, (byte)alpha);
                imgData.put(i, (byte)col.getRed());
                imgData.put(i+1, (byte)col.getGreen());
                imgData.put(i+2, (byte)col.getBlue());
                tex.texData[j] = new Color(col.getRed(), col.getGreen(), col.getBlue(), alpha);
            }
        }
	
        GL11.glGenTextures(tmp);
        GL11.glBindTexture(GL11.GL_TEXTURE_2D, tmp.get(0));
        GLU.gluBuild2DMipmaps(GL11.GL_TEXTURE_2D, 4, tex.texWidth, tex.texHeight, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, imgData);
        //GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 4, tex.texWidth,tex.texHeight, 0,
		//	    GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, imgData);
        if(linearFilter)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        }
        else
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
        }
        
        if(clampEdges)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_S, GL11.GL_CLAMP);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_WRAP_T, GL11.GL_CLAMP);
        }
	tex.textureID = tmp.get(0);
	
	return tex;
    }
    
    public static Texture createFromRawData(ByteBuffer buf, int w, int h, int bpp, boolean linearFilter)
    {
        Texture tex = new Texture();
        tex.hasAlpha = (bpp == 4);
        tex.realWidth = w;
        tex.realHeight = h;
        tex.texWidth = w;
        tex.texHeight = h;
        tex.depth = bpp;
        
        IntBuffer tmp = BufferUtils.createIntBuffer(1);
        GL11.glGenTextures(tmp);
        GL11.glBindTexture(GL11.GL_TEXTURE_2D, tmp.get(0));
	
        if(bpp == 3)
            GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 3, tex.texWidth,tex.texHeight, 0,
                              GL11.GL_RGB, GL11.GL_UNSIGNED_BYTE, buf);
        else if(bpp == 4)
            GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, 4, tex.texWidth,tex.texHeight, 0,
                              GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buf);
	
        if(linearFilter)
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_LINEAR);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_LINEAR);
        }
        else
        {
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
        }
        tex.textureID = tmp.get(0);

        return tex;
    }
    
    public Color getPixel(int x, int y)
    {	
	if(x < 0 || y < 0 || x >= realWidth || y >= realHeight)
	    return new Color(0,0,0,0);
	int i = (y*realWidth + x);
	
	Color col = texData[i];
	return col;
    }
    
    public boolean hasAlpha()
    {
        return hasAlpha;
    }
}
