/*
 * Map.java
 *
 * Created: September 29, 2011
 * Author: Stephen Whitmore
 *
 */
package irrtum.map;

import irrtum.game.*;
import irrtum.engine.*;
import java.util.HashMap;
import java.util.Random;
import org.lwjgl.opengl.*;


public class Map
{
    protected HashMap<String,Chunk> chunks;

    public Map()
    {
        chunks = new HashMap<String,Chunk>();

        chunks.put(createCanonicalChunkString(0, 0, 0), new Chunk(this, 0,0,0));
        chunks.put(createCanonicalChunkString(1, 0, 0), new Chunk(this, 1,0,0));
        chunks.put(createCanonicalChunkString(2, 0, 0), new Chunk(this, 2,0,0));
        chunks.put(createCanonicalChunkString(0, 1, 0), new Chunk(this, 0,1,0));
        chunks.put(createCanonicalChunkString(1, 1, 0), new Chunk(this, 1,1,0));
        chunks.put(createCanonicalChunkString(2, 1, 0), new Chunk(this, 2,1,0));

        generateLevel();
        for(Chunk chunk : chunks.values())
        {
            chunk.fixGrass();
        }
        calculateLighting();
    }

    public void generateLevel()
    {
        BlockDef grass = new BlockDef(true, true);
        grass.setFaceIndices(0, 2, 1, 1, 1, 1);
        BlockDef dirt = new BlockDef(true, true);
        dirt.setFaceIndices(2, 2, 2, 2, 2, 2);
        Random rand = new Random();

        for(int k=0; k < 20; k++)
        {
            float radius = rand.nextInt(4)+4;
            Vector3 pos = new Vector3(rand.nextInt(Chunk.CHUNK_SIZE*2+20)-10,rand.nextInt(Chunk.CHUNK_SIZE*2+20)-10,rand.nextInt(Chunk.CHUNK_SIZE*2+20)-10);
            Vector3 dir = Vector3.random();
            for(int i=0; i < 50; i++)
            {
                this.carveSphere((int)pos.x, (int)pos.y, (int)pos.z, (int)radius);
                pos = pos.add(dir);
                dir = dir.add(Vector3.random()).norm();
                radius += rand.nextFloat() * 2.0f - 1.0f;
            }
        }
    }

    public void render(Camera cam)
    {
        GL11.glLoadIdentity();

        cam.doTransform();
        Sun.setupLight();

        for(Chunk chunk : chunks.values())
        {
            GL11.glPushMatrix();
            GL11.glTranslatef(chunk.getX()*Chunk.CHUNK_SIZE*Block.BLOCK_SIZE,
                              chunk.getY()*Chunk.CHUNK_SIZE*Block.BLOCK_SIZE,
                              chunk.getZ()*Chunk.CHUNK_SIZE*Block.BLOCK_SIZE);
            chunk.render();
            GL11.glPopMatrix();
        }
    }

    public Chunk getChunkAt(int x, int y, int z)
    {
        int chunkX = x / Chunk.CHUNK_SIZE;
        int chunkY = y / Chunk.CHUNK_SIZE;
        int chunkZ = z / Chunk.CHUNK_SIZE;
        String key = createCanonicalChunkString(chunkX, chunkY, chunkZ);

        return chunks.get(key);
    }

    public Block getBlockAt(int x, int y, int z)
    {
        Chunk chunk = getChunkAt(x,y,z);

        if(chunk != null)
        {
            int localX = x - chunk.getX() * Chunk.CHUNK_SIZE;
            int localY = y - chunk.getY() * Chunk.CHUNK_SIZE;
            int localZ = z - chunk.getZ() * Chunk.CHUNK_SIZE;

            return chunk.getBlockAt(localX, localY, localZ);
        }
        else
        {
            return null;
        }
    }

    public void clearBlockAt(int x, int y, int z)
    {
        Chunk chunk = getChunkAt(x,y,z);

        if(chunk != null)
        {
            int localX = x - chunk.getX() * Chunk.CHUNK_SIZE;
            int localY = y - chunk.getY() * Chunk.CHUNK_SIZE;
            int localZ = z - chunk.getZ() * Chunk.CHUNK_SIZE;

            chunk.clearBlockAt(localX, localY, localZ);
        }
    }

    public void calculateLighting()
    {
        for(Chunk chunk : chunks.values())
        {
            chunk.calculateLighting();
        }
    }

    private static String createCanonicalChunkString(int x, int y, int z)
    {
        StringBuilder strBuilder = new StringBuilder();
        strBuilder.append(x);
        strBuilder.append(",");
        strBuilder.append(y);
        strBuilder.append(",");
        strBuilder.append(z);

        return strBuilder.toString();
    }

    public void carveSphere(int x, int y, int z, int radius)
    {
        for(int i=-x-radius; i < x+radius; i++)
        {
            for(int j=-y-radius; j < y+radius; j++)
            {
                for(int k=-z-radius; k < z+radius; k++)
                {
                    float dist2 = (i-x)*(i-x) + (j-y)*(j-y) + (k-z)*(k-z);
                    if(dist2 <= radius*radius)
                    {
                        clearBlockAt(i,j,k);
                    }
                }
            }
        }
    }
}




class Chunk
{
    // Size of each chunk (in units of BLOCKSxBLOCKS)
    public static final int CHUNK_SIZE = 25;

    Map m_Map;
    Block[][][] blocks;
    int x, y, z;

    public Chunk(Map parentMap, int xo, int yo, int zo)
    {
        m_Map = parentMap;

        BlockDef grass = new BlockDef(true, true);
        grass.setFaceIndices(0, 2, 1, 1, 1, 1);
        BlockDef dirt = new BlockDef(true, true);
        dirt.setFaceIndices(2, 2, 2, 2, 2, 2);

        x = xo; y = yo; z = zo;
        blocks = new Block[CHUNK_SIZE][CHUNK_SIZE][CHUNK_SIZE];

        for(int i=0; i < CHUNK_SIZE; i++)
        {
            for(int j=0; j < CHUNK_SIZE; j++)
            {
                for(int k=0; k < CHUNK_SIZE; k++)
                {
                    blocks[i][j][k] = new Block(dirt);
                    /*
                    if(j%CHUNK_SIZE==0 || i%(CHUNK_SIZE-1)==0 || k%(CHUNK_SIZE-1)==0)
                    {
                        if(j==0 || j==CHUNK_SIZE/(xo+1)-1)
                        {
                            blocks[i][j][k] = new Block(grass);
                        }
                        else
                        {
                            blocks[i][j][k] = new Block(dirt);
                        }
                    }
                    if(i >= 15 && i <= 20 && k >= 10 && k <= 20 && j ==CHUNK_SIZE/(xo+1)-1)
                    {
                        blocks[i][j][k] = new Block(grass);
                    }
                    */
                }
            }
        }
    }

    public void fixGrass()
    {
        BlockDef grass = new BlockDef(true, true);
        grass.setFaceIndices(0, 2, 1, 1, 1, 1);
        BlockDef dirt = new BlockDef(true, true);
        dirt.setFaceIndices(2, 2, 2, 2, 2, 2);

        for(int i=0; i < CHUNK_SIZE; i++)
        {
            for(int j=0; j < CHUNK_SIZE; j++)
            {
                for(int k=0; k < CHUNK_SIZE; k++)
                {
                    Block block = getBlockAt(i,j,k);
                    if(block != null)
                    {
                        if(m_Map.getBlockAt(x*CHUNK_SIZE+i, y*CHUNK_SIZE+j+1, z*CHUNK_SIZE+k) == null)
                        {
                            block.setBlockDef(grass);
                        }
                        else
                        {
                            block.setBlockDef(dirt);
                        }
                    }
                }
            }
        }
    }

    public void render()
    {
        Texture texture = ResourceManager.getImage("Blocks").getTexture();
        texture.bind();
        int facesWidth = (int)(texture.getWidth() / 16.0f);
        int facesHeight = (int)(texture.getHeight() / 16.0f);
        float uSize = 1.0f / (float)facesWidth;
        float vSize = 1.0f / (float)facesHeight;

        GL11.glPushMatrix();
        GL11.glBegin(GL11.GL_QUADS);
        for(int i=0; i < CHUNK_SIZE; i++)
        {
            for(int j=0; j < CHUNK_SIZE; j++)
            {
                for(int k=0; k < CHUNK_SIZE; k++)
                {
                    float xo = i * Block.BLOCK_SIZE;
                    float yo = j * Block.BLOCK_SIZE;
                    float zo = k * Block.BLOCK_SIZE;

                    Block block = getBlockAt(i, j, k);
                    if(block != null)
                    {
                        Block above = getBlockAt(i, j+1, k);
                        if(above == null)
                        {
                            GL11.glColor3f(block.getBrightness(), block.getBrightness(), block.getBrightness());
                            float u = (block.getBlockDef().getFaceIndex(0) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(0) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(0, 1, 0);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo);
                        }

                        float otherBrightness = Math.min(0.75f, block.getBrightness());

                        Block below = getBlockAt(i, j-1, k);
                        if(below == null)
                        {
                            GL11.glColor3f(otherBrightness*0.75f, otherBrightness*0.75f, otherBrightness*0.75f);
                            float u = (block.getBlockDef().getFaceIndex(1) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(1) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(0, -1, 0);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo,                     yo,                  zo);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo,                     yo,                  zo+Block.BLOCK_SIZE);
                        }
                        Block front = getBlockAt(i, j, k+1);
                        if(front == null)
                        {
                            GL11.glColor3f(otherBrightness*0.95f, otherBrightness*0.95f, otherBrightness*0.95f);
                            float u = (block.getBlockDef().getFaceIndex(2) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(2) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(0, 0, 1);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo,                     yo,                  zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                        }
                        Block behind = getBlockAt(i, j, k-1);
                        if(behind == null)
                        {
                            GL11.glColor3f(otherBrightness*0.9f, otherBrightness*0.9f, otherBrightness*0.9f);
                            float u = (block.getBlockDef().getFaceIndex(3) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(3) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(0, 0, -1);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo,                     yo,                  zo);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo);
                        }
                        Block left = getBlockAt(i-1, j, k);
                        if(left == null)
                        {
                            GL11.glColor3f(otherBrightness*0.85f, otherBrightness*0.85f, otherBrightness*0.85f);
                            float u = (block.getBlockDef().getFaceIndex(4) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(4) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(-1, 0, 0);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo,                     yo,                  zo);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo,                     yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo,                     yo,                  zo+Block.BLOCK_SIZE);
                        }
                        Block right = getBlockAt(i+1, j, k);
                        if(right == null)
                        {
                            GL11.glColor3f(otherBrightness*0.8f, otherBrightness*0.8f, otherBrightness*0.8f);
                            float u = (block.getBlockDef().getFaceIndex(5) % facesWidth) / (float)facesWidth;
                            float v = (block.getBlockDef().getFaceIndex(5) / facesWidth) / (float)facesHeight;
                            GL11.glNormal3f(1, 0, 0);
                            GL11.glTexCoord2f(u,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo);
                            GL11.glTexCoord2f(u+uSize,v+vSize);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo,                  zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u+uSize,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo+Block.BLOCK_SIZE);
                            GL11.glTexCoord2f(u,v);
                            GL11.glVertex3f(xo+Block.BLOCK_SIZE,    yo+Block.BLOCK_SIZE, zo);
                        }
                    }
                }
            }
        }
        GL11.glEnd();
        GL11.glPopMatrix();
        GL11.glColor3f(1, 1, 1);
    }

    protected void calculateLighting()
    {
        for(int i=0; i < CHUNK_SIZE; i++)
        {
            for(int j=0; j < CHUNK_SIZE; j++)
            {
                for(int k=0; k < CHUNK_SIZE; k++)
                {
                    Block block = getBlockAt(i, j, k);
                    if(block == null)
                    {
                        continue;
                    }

                    Vector3 dir = Sun.dir.scale(4.0f);
                    dir = new Vector3(0.0f, -1.0f, 0.0f).norm().scale(4.0f);
                    Vector3 pos = new Vector3(x * CHUNK_SIZE * Block.BLOCK_SIZE,
                                              y * CHUNK_SIZE * Block.BLOCK_SIZE,
                                              z * CHUNK_SIZE * Block.BLOCK_SIZE);
                    pos = pos.add(new Vector3(i * Block.BLOCK_SIZE + Block.BLOCK_SIZE/2,
                                              j * Block.BLOCK_SIZE + Block.BLOCK_SIZE/2,
                                              k * Block.BLOCK_SIZE + Block.BLOCK_SIZE/2));

                    for(int m=0; m < 100; m++)
                    {
                        pos = pos.sub(dir);
                        Block hitBlock = m_Map.getBlockAt( (int)(pos.x/(float)Block.BLOCK_SIZE),
                                                           (int)(pos.y/(float)Block.BLOCK_SIZE),
                                                           (int)(pos.z/(float)Block.BLOCK_SIZE));
                        if(hitBlock != null && hitBlock != block)
                        {
                            block.setBrightness(0.5f);
                            break;
                        }
                    }
                }
            }
        }
    }

    public Block getBlockAt(int x, int y, int z)
    {
        if(x < 0 || y < 0 || z < 0 || x >= CHUNK_SIZE || y >= CHUNK_SIZE || z >= CHUNK_SIZE)
        {
            return null;
        }
        else
        {
            return blocks[x][y][z];
        }
    }

    void clearBlockAt(int localX, int localY, int localZ)
    {
        if(localX < 0 || localY < 0 || localZ < 0 || localX >= CHUNK_SIZE || localY >= CHUNK_SIZE || localZ >= CHUNK_SIZE)
        {
            return;
        }

        blocks[localX][localY][localZ] = null;
    }

    public int getX()
    {
        return x;
    }
    public int getY()
    {
        return y;
    }
    public int getZ()
    {
        return z;
    }
}


class Block
{
    public static int BLOCK_SIZE = 16;

    public BlockDef parentDef;
    private float m_Brightness;

    public Block(BlockDef def)
    {
        parentDef = def;
        m_Brightness = 1.0f;
    }

    public BlockDef getBlockDef()
    {
        return parentDef;
    }
    public void setBlockDef(BlockDef def)
    {
        parentDef = def;
    }

    public float getBrightness()
    {
        return m_Brightness;
    }
    public void setBrightness(float brightness)
    {
        m_Brightness = brightness;
    }
}


class BlockDef
{
    public boolean m_IsOpaque;
    public boolean m_IsSolid;
    protected int[] m_FaceTextures;

    public BlockDef(boolean opaque, boolean solid)
    {
        //m_Texture = ResourceManager.getImage(texture).getTexture();
        m_IsOpaque = opaque;
        m_IsSolid = solid;
        m_FaceTextures = new int[6];
        for(int i=0; i < 6; i++) { m_FaceTextures[i] = 0; }
    }

    public void setFaceIndices(int top, int bottom, int front, int back, int left, int right)
    {
        m_FaceTextures[0] = top;
        m_FaceTextures[1] = bottom;
        m_FaceTextures[2] = front;
        m_FaceTextures[3] = back;
        m_FaceTextures[4] = left;
        m_FaceTextures[5] = right;
    }

    public int getFaceIndex(int i)
    {
        return m_FaceTextures[i];
    }
}
